import manim
from manim.constants import *

from ..useful_in_development import *
import numpy as np


class Line(manim.Line):
    def __init__(self, start=LEFT, end=manim.RIGHT, auto_dot_to_array=True, **kwargs):
        if auto_dot_to_array:
            start, end = dot_to_array(start, end)
        super().__init__(start, end, **kwargs)

    def is_point_in_line(self, dot):
        dot = dot_to_array(dot)[0]
        x1, y1 = self.get_start()[0], self.get_start()[1]
        x2, y2 = self.get_end()[0], self.get_end()[1]
        x3, y3 = dot[0], dot[1]
        return round(np.linalg.det(np.array([[x1, y1, 1], [x2, y2, 1], [x3, y3, 1]])), 3) == 0

    def set_length_about_point(self, length, dot):
        dot = dot_to_array(dot)[0]
        A = dot + np.array([length, 0, 0])
        line = Line(dot, A).rotate(about_point=dot, angle=self.get_angle())
        self.put_start_and_end_on(line.get_start(), line.get_end())
        return self
    
    def get_distance(self, dot):
        dot = dot_to_array(dot)[0]
        return Line(dot, self.get_projection(dot)).get_length()

    def equal(self, n=1, buff=0.1, length=0.2, **kwargs):
        return manim.VGroup(*[Line(**kwargs).set_length(length).rotate(
            about_point=Line(**kwargs).set_length(length).get_center(), angle=np.pi / 2) for i in range(n)]).arrange(
            buff=buff, direction=RIGHT).move_to(self.get_center()).rotate(
            about_point=self.get_center(), angle=self.get_angle()).set_z_index(1)

    def paral(self, n=1, buff=0.08, length=0.15, rotate=False, **kwargs):
        elem = manim.VGroup(
            Line(**kwargs).set_length(length).rotate(about_point=Line().set_length(0.2).get_end(),
                                                     angle=PI / 4).shift(0.02 * UR),
            Line(**kwargs).set_length(length).rotate(about_point=Line().set_length(0.2).get_end(),
                                                     angle=-PI / 4).shift(0.02 * DR)
        ).rotate(int(rotate) * np.pi)
        return manim.VGroup(*[elem.copy() for i in range(n)]).arrange(buff=buff, direction=RIGHT).move_to(
            self.get_center()).rotate(about_point=self.get_center(), angle=self.get_angle()).set_z_index(1)

    def inversion(self, circle, **kwargs):
        if self.is_point_in_line(circle.get_center()):
            return Line(self.get_start(), self.get_end(), **kwargs)
        else:
            dot_1 = circle_symmetry(circle, self.point_from_proportion(0.3))
            dot_2 = circle_symmetry(circle, self.point_from_proportion(0.5))
            dot_3 = circle_symmetry(circle, self.point_from_proportion(0.7))
            return Circle().from_three_points(dot_1, dot_2, dot_3, **kwargs)


class Angle(manim.Angle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def from_three_points(A, B, C, auto_dot_to_array=True, **kwargs):
        angle = Angle(Line(B, A), Line(B, C), **kwargs)
        if angle.get_angle_value() > PI:
            angle = Angle(Line(B, C), Line(B, A), **kwargs)
        return angle

    def get_label_center(self, delta=0.25):
        A = self.lines[0].get_end()
        B = self.lines[0].get_start()
        C = self.lines[1].get_end()
        A, B, C = dot_to_array(A, B, C)

        try:
            radius = self.radius
        except AttributeError:
            inter = manim.line_intersection(
                [self.lines[0].get_start(), self.lines[0].get_end()],
                [self.lines[1].get_start(), self.lines[1].get_end()],
            )

            if self.quadrant[0] == 1:
                dist_1 = np.linalg.norm(self.lines[0].get_end() - inter)
            else:
                dist_1 = np.linalg.norm(self.lines[0].get_start() - inter)
            if self.quadrant[1] == 1:
                dist_2 = np.linalg.norm(self.lines[1].get_end() - inter)
            else:
                dist_2 = np.linalg.norm(self.lines[1].get_start() - inter)
            if np.minimum(dist_1, dist_2) < 0.6:
                radius = (2 / 3) * np.minimum(dist_1, dist_2)
            else:
                radius = 0.4

        center = Line(B, Line(A, C).point_from_proportion(((Line(A, C).get_length() * Line(A, B).get_length()) / (
                Line(B, C).get_length() + Line(A, B).get_length())) / Line(A, C).get_length())).set_length_about_point(
            delta + radius, B).get_end()

        return center



class Circle(manim.Circle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def from_three_points(p1, p2, p3, auto_dot_to_array=True, **kwargs):
        if auto_dot_to_array:
            p1, p2, p3 = dot_to_array(p1, p2, p3)
        center = manim.line_intersection(
            manim.perpendicular_bisector([p1, p2]),
            manim.perpendicular_bisector([p2, p3]),
        )
        radius = np.linalg.norm(p1 - center)
        return Circle(radius=radius, **kwargs).shift(center)

    def pow(self, dot):
        dot = dot_to_array(dot)[0]
        return round((pow(self.get_center()[0] - dot[0], 2) + pow(self.get_center()[1] - dot[1], 2)) -
                     self.radius ** 2, 4)

    def is_point_in_circle(self, dot):
        dot = dot_to_array(dot)[0]
        return self.pow(dot) == 0


class Dot(manim.Dot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from __future__ import annotations

import numpy as np

from .default import Line, Circle, Dot, Angle
from .intersection import intersection_circles
from ..useful_in_development import *
from .triangle_centers import Incenter


__all__ = [
    "Tangent",
    "Incircle",

]

class Tangent(Line):
    def __init__(self, circle: Circle, dot: Dot, is_other_tangent=False, **kwargs):
        dot = dot_to_array(dot)[0]
        if (circle.pow(dot) < 0):
            raise Exception("Dot is inside the circle")

        self.circle = circle
        self.dot = dot
        self.tangent_point = Dot(intersection_circles(
            circle, Circle(arc_center=dot, radius=np.sqrt(
                pow(distance(circle.get_center(), dot), 2) - pow(circle.get_radius(), 2)
            ))
        )[is_other_tangent])
        self.is_other_tangent = is_other_tangent
        super().__init__(dot, self.tangent_point, **kwargs)

    def get_other_tangent(self):
        return Tangent(self.circle, self.dot, is_other_tangent=(not self.is_other_tangent))


class Incircle(Circle):
    def __init__(self, A, B, C, *args, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        I = Incenter(A, B, C).get_center()
        a = distance(B, C)
        b = distance(A, C)
        c = distance(A, B)
        p = (a + b + c) / 2
        r = np.sqrt((p - a) * (p - b) * (p - c) / p)
        super().__init__(radius=r, arc_center=I, *args, **kwargs)
        
        







import numpy as np
import manim


def dot_to_array(*dots) -> list:
    coordinates = []
    for dot in dots:
        if isinstance(dot, np.ndarray):
            coordinates.append(dot)
        else:
            try:
                coordinates.append(dot.get_center())
            except:
                pass
    return coordinates


def circle_symmetry(circle: manim.Circle, dot) -> np.ndarray:
    x, y, z = dot_to_array(dot)
    x0, y0, z0 = circle.get_center()
    r = circle.get_radius()
    return np.array([
        x0 + ((r ** 2 * (x - x0)) / ((x - x0) ** 2 + (y - y0) ** 2)),
        y0 + ((r ** 2 * (y - y0)) / ((x - x0) ** 2 + (y - y0) ** 2)),
        0
    ])


def distance(A, B) -> float:
    A, B = dot_to_array(A, B)
    return np.sqrt((A[0] - B[0])**2 + (A[1] - B[1])**2 + (A[2] - B[2])**2)




from __future__ import annotations

from ..useful_in_development import *

from .default import Line, Angle
from manim import VGroup
from manim.constants import PI

__all__ = [
    "Cevian",
    "Bisector",
    "Median",
    "Perpendicular",
    "Altitude",
    "PerpendicularBisector",
    "Symmedian"
]


class Cevian(Line):
    def __init__(self, A, B, C, alpha=0.35, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        D = Line(A, C).point_from_proportion(alpha)

        self.foot = D
        self.general_vertex = B
        self.extra_vertex_1 = A
        self.extra_vertex_2 = C
        self.proportion = alpha

        super().__init__(B, D, **kwargs)

    def get_foot(self) -> np.ndarray:
        return self.foot

    def get_vertex(self) -> np.ndarray:
        return self.general_vertex

    def get_proportion(self) -> float:
        return self.proportion


class Bisector(Cevian):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        alpha = ((Line(A, C).get_length() * Line(A, B).get_length()) / (
                    Line(B, C).get_length() + Line(A, B).get_length())) / Line(A, C).get_length()
        super().__init__(A, B, C, alpha, **kwargs)

    def bisected_angles(self, **kwargs) -> VGroup(Angle, Angle):
        return VGroup(
            Angle().from_three_points(self.extra_vertex_1, self.general_vertex, self.foot, **kwargs),
            Angle().from_three_points(self.extra_vertex_2, self.general_vertex, self.foot, **kwargs)
        )


class Median(Cevian):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        super().__init__(A, B, C, 0.5, **kwargs)

    def equal_segments(self, **kwargs) -> VGroup(Line, Line):
        return VGroup(
            Line(self.extra_vertex_1, self.foot).equal(**kwargs),
            Line(self.extra_vertex_2, self.foot).equal(**kwargs),
        )


class Symmedian(Cevian):
    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        a, b, c = Line(B, C).get_length(), Line(A, C).get_length(), Line(A, B).get_length()
        alpha = c**2 / (a**2 + c**2)
        super().__init__(A, B, C, alpha=alpha, **kwargs)


class Perpendicular(Line):
    def __init__(self, line: Line, dot, length=1.0, rotate=False, **kwargs):
        X = dot_to_array(dot)[0]
        A, B = line.get_start(), line.get_end()
        if Line(A, B).is_point_in_line(X):
            perpendicular = Line(A, X).rotate(
                about_point=X, angle=PI / 2 + int(rotate)*PI).set_length_about_point(length, X)
        else:
            perpendicular = Line(X, Line(A, B).get_projection(X))

        self.vertex = X
        self.foot = Line(A, B).get_projection(X)
        self.line = line

        super().__init__(perpendicular.get_start(), perpendicular.get_end(), **kwargs)

    def get_line(self) -> Line:
        return self.linez

    def get_vertex(self) -> np.ndarray:
        return self.vertex

    def get_foot(self) -> np.ndarray:
        return self.foot

    def get_angles(self, **kwargs):
        return VGroup(
            Angle().from_three_points(self.vertex, self.dot, self.line.get_start(), elbow=True, **kwargs),
            Angle().from_three_points(self.vertex, self.dot, self.line.get_end(), elbow=True, **kwargs)
        )


class PerpendicularBisector(Line):
    def __init__(self, line: Line, length=1, **kwargs):
        line_1 = Perpendicular(line=line, dot=line.get_center(), length=length / 2, rotate=False)
        line_2 = Perpendicular(line=line, dot=line.get_center(), length=length / 2, rotate=True)

        self.dot = line.get_center()
        self.line = line

        super().__init__(line_1.get_end(), line_2.get_end(), **kwargs)

    def get_angles(self, **kwargs):
        return VGroup(
            Angle().from_three_points(self.line.get_start(), self.dot, self.get_start(), **kwargs),
            Angle().from_three_points(self.line.get_start(), self.dot, self.get_end(), **kwargs),
            Angle().from_three_points(self.line.get_end(), self.dot, self.get_start(), **kwargs),
            Angle().from_three_points(self.line.get_end(), self.dot, self.get_end(), **kwargs),
        )


class Altitude(Perpendicular):

    def __init__(self, A, B, C, **kwargs):
        A, B, C = dot_to_array(A, B, C)
        line = Line(A, C)
        super().__init__(line, B, **kwargs)



from __future__ import annotations

import manim
from manim.constants import *

__all__ = [
    "Fancy_label"
]


def Fancy_label(text: manim.Tex, mode='normal', buff=manim.MED_LARGE_BUFF):
    mode = mode.lower()
    modes = ['vlow', 'low', 'normal', 'fast', 'vfast']

    if mode not in modes:
        mode = 'normal'


    symbols = len(text.get_tex_string())
    time = (symbols / (9+modes.index(mode))) + 0.25

    return manim.AnimationGroup(manim.Write(text.to_edge(UP, buff=buff).set_z_index(999)), run_time=time)